#define IMPORT_CORE
#include "core/WindowManager.h"

#define GLFW_INCLUDE_VULKAN
#include <GLFW/glfw3.h>
#include <core/HashMap.h>
#include <core/Logger.h>
#include <core/Utility.h>
#include <limits.h>
#include <stdio.h>
#include <uchar.h>

static bool dummyWindowRunHandler(void*) {
    return !shouldWindowClose();
}

static void dummyWindowTickHandler(void*) {
}

static void dummyWindowRenderHandler(void*, float) {
}

static GLFWwindow* window = nullptr;
static IntVector2 size = {0};
static bool sizeChanged = false;
static WindowRunHandler runHandler = dummyWindowRunHandler;
static WindowTickHandler tickHandler = dummyWindowTickHandler;
static WindowRenderHandler renderHandler = dummyWindowRenderHandler;
static void* runHandlerData = nullptr;
static void* tickHandlerData = nullptr;
static void* renderHandlerData = nullptr;
static i64 nanosPerTick = 1;

#define CLOCK_SIZE 16
typedef struct {
    i64 last;
    i64 sum;
    i64 values[CLOCK_SIZE];
    size_t index;
} Clock;
static Clock fps = {0};
static Clock tps = {0};

#define INPUT_SIZE 2048
static char input[INPUT_SIZE] = {0};
static size_t inputLength = 0;
static size_t inputCursor = 0;
static size_t inputLimit = INPUT_SIZE - 1;
static bool inputActive = false;

typedef struct {
    char name[31];
    bool released;
    int key;
    int downTime;
    int upEvents;
    int downEvents;
    int controllerUp;
    int controllerDown;
} ButtonData;
#define BUTTONS 100
static ButtonData buttons[BUTTONS] = {{.name = "unknown"}};
static size_t buttonIndex = 1;
#define KEYS (GLFW_KEY_LAST + 1)
static Button keyToButton[KEYS] = {0};
#define GAMEPAD_BUTTONS (GLFW_GAMEPAD_BUTTON_LAST + 1)
static Button gamepadToButton[GAMEPAD_BUTTONS] = {0};
#define MOUSE_BUTTONS (GLFW_MOUSE_BUTTON_LAST + 1)
static Button mouseToButton[MOUSE_BUTTONS] = {0};
static Vector2 lastMousePosition = {0};
static Vector2 mousePosition = {0};
static int activeController = -1;
static GLFWgamepadstate lastControllerState = {0};

static void onButton(Button* map, int n, int key, int action) {
    if(key < 0 || key >= n) {
        return;
    }
    Button b = map[key];
    if(b == 0 || b >= buttonIndex) {
        return;
    }
    if(action == GLFW_RELEASE) {
        buttons[b].upEvents++;
    } else if(action == GLFW_PRESS) {
        buttons[b].downEvents++;
    }
}

static bool isCharSequence(size_t i) {
    return (input[i] & 0xC0) == 0x80;
}

static void handleInputKey(int key, int action) {
    if(action == GLFW_RELEASE) {
        return;
    }
    switch(key) {
        case GLFW_KEY_BACKSPACE:
            if(inputLength >= inputCursor && inputCursor > 0) {
                size_t w = 1;
                while(inputCursor - w > 0 && isCharSequence(inputCursor - w)) {
                    w++;
                }
                for(size_t i = inputCursor - w; i < inputLength; i++) {
                    input[i] = input[i + w];
                }
                inputLength -= w;
                inputCursor -= w;
            }
            break;
        case GLFW_KEY_LEFT: {
            inputCursor -= inputCursor > 0;
            while(inputCursor > 0 && isCharSequence(inputCursor)) {
                inputCursor--;
            }
            break;
        }
        case GLFW_KEY_RIGHT: {
            inputCursor += inputCursor < inputLength;
            while(inputCursor < inputLength && isCharSequence(inputCursor)) {
                inputCursor++;
            }
            break;
        }
    }
}

static void onKey(GLFWwindow*, int key, int scancode, int action, int mods) {
    (void)scancode;
    (void)mods;
    if(inputActive) {
        handleInputKey(key, action);
    }
    onButton(keyToButton, KEYS, key, action);
}

static void addUnicode(u32 c) {
    char buffer[MB_LEN_MAX + 1] = {0};
    mbstate_t state = {0};
    size_t w = c32rtomb(buffer, c, &state);
    if(w >= MB_LEN_MAX || inputLength + w >= inputLimit) {
        return;
    }
    if(inputLength > 0) {
        for(size_t i = inputLength - 1; i >= inputCursor; i--) {
            swap(input + (i + w), input + i);
        }
    }
    memcpy(input + inputCursor, buffer, w);
    inputLength += w;
    inputCursor += w;
}

static void onChar(GLFWwindow*, u32 codepoint) {
    if(inputActive) {
        addUnicode(codepoint);
    }
}

static void onResize(GLFWwindow*, int width, int height) {
    sizeChanged = true;
    size.data[0] = width;
    size.data[1] = height;
}

static void onMouse(GLFWwindow*, int button, int action, int mods) {
    (void)mods;
    onButton(mouseToButton, MOUSE_BUTTONS, button, action);
}

static void onMouseMove(GLFWwindow*, double x, double y) {
    mousePosition.data[0] = (float)x;
    mousePosition.data[1] = (float)y;
}

#define VK_ERROR_CASE(error)                                                   \
    case error: return #error

static const char* getVulkanErrorString(VkResult r) {
    switch(r) {
        VK_ERROR_CASE(VK_ERROR_OUT_OF_HOST_MEMORY);
        VK_ERROR_CASE(VK_ERROR_OUT_OF_DEVICE_MEMORY);
        VK_ERROR_CASE(VK_ERROR_INITIALIZATION_FAILED);
        VK_ERROR_CASE(VK_ERROR_LAYER_NOT_PRESENT);
        VK_ERROR_CASE(VK_ERROR_EXTENSION_NOT_PRESENT);
        VK_ERROR_CASE(VK_ERROR_INCOMPATIBLE_DRIVER);
        default: return "unknown";
    }
}

#define VK_ASSERT(a)                                                           \
    do {                                                                       \
        VkResult vkResult = (a);                                               \
        if(vkResult != VK_SUCCESS) {                                           \
            printf("Vulkan error at [%s:%d]: %s\n", __FUNCTION__, __LINE__,    \
                   getVulkanErrorString(vkResult));                            \
            return true;                                                       \
        }                                                                      \
    } while(false)

static VkResult findSuitablePhysicalDevice(VkInstance instance,
                                           VkPhysicalDevice* pd) {
    u32 deviceCount = 0;
    VkResult r = vkEnumeratePhysicalDevices(instance, &deviceCount, nullptr);
    if(r != VK_SUCCESS) {
        return r;
    } else if(deviceCount == 0) {
        return VK_ERROR_INITIALIZATION_FAILED;
    }
    VkPhysicalDevice* devices =
        cAllocate(sizeof(VkPhysicalDevice) * deviceCount);
    r = vkEnumeratePhysicalDevices(instance, &deviceCount, devices);
    if(r != VK_SUCCESS) {
        cFree(devices);
        return r;
    }
    int bestPoints = 0;
    LOG_INFO("Found %u devices", deviceCount);
    for(u32 i = 0; i < deviceCount; i++) {
        int points = 0;
        VkPhysicalDeviceProperties p;
        vkGetPhysicalDeviceProperties(devices[i], &p);
        if(p.deviceType == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU) {
            points += 100;
        } else if(p.deviceType == VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU) {
            points += 50;
        } else if(p.deviceType == VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU) {
            points += 20;
        }
        if(points > bestPoints) {
            bestPoints = points;
            *pd = devices[i];
            LOG_INFO("Best Device: %s", p.deviceName);
        }
    }
    cFree(devices);
    return bestPoints > 0 ? VK_SUCCESS : VK_ERROR_INITIALIZATION_FAILED;
}

static u32 findQueueFamilies(VkPhysicalDevice pd, VkQueueFlags flags) {
    u32 count = 0;
    vkGetPhysicalDeviceQueueFamilyProperties(pd, &count, nullptr);
    VkQueueFamilyProperties* properties =
        cAllocate(sizeof(VkQueueFamilyProperties) * count);
    vkGetPhysicalDeviceQueueFamilyProperties(pd, &count, properties);
    for(u32 i = 0; i < count; i++) {
        if(properties[i].queueCount != 0 &&
           (properties[i].queueFlags & flags) == flags) {
            LOG_INFO("Queue family bits: %x", properties[i].queueFlags);
            cFree(properties);
            return i;
        }
    }
    cFree(properties);
    return (u32)-1;
}

bool openWindow(const WindowOptions* o) {
    if(!glfwInit()) {
        LOG_ERROR("could not initialize GLFW");
        return true;
    }

    glfwDefaultWindowHints();
    glfwWindowHint(GLFW_VISIBLE, false);
    glfwWindowHint(GLFW_RESIZABLE, true);
    glfwWindowHint(GLFW_DECORATED, !o->fullscreen);
    glfwWindowHint(GLFW_DOUBLEBUFFER, true);
    glfwWindowHint(GLFW_CLIENT_API, GLFW_NO_API);

    GLFWmonitor* m = o->fullscreen ? glfwGetPrimaryMonitor() : nullptr;
    window =
        glfwCreateWindow(o->size.data[0], o->size.data[1], o->name, m, nullptr);
    if(window == nullptr) {
        LOG_ERROR("could not create window");
        closeWindow();
        return true;
    }
    size = o->size;
    glfwSetKeyCallback(window, onKey);
    glfwSetCharCallback(window, onChar);
    glfwSetFramebufferSizeCallback(window, onResize);
    glfwSetMouseButtonCallback(window, onMouse);
    glfwSetCursorPosCallback(window, onMouseMove);

    VkInstance instance = {0};
    VkApplicationInfo info = {.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO,
                              .pApplicationName = "Vulkan",
                              .applicationVersion = VK_MAKE_VERSION(1, 0, 0),
                              .pEngineName = "Kajetan",
                              .engineVersion = VK_MAKE_VERSION(0, 0, 1),
                              .apiVersion = VK_API_VERSION_1_1};
    VkInstanceCreateInfo instanceCreateInfo = {
        .sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
        .pApplicationInfo = &info,
        .enabledLayerCount = 1,
        .ppEnabledLayerNames = (const char*[]){"VK_LAYER_KHRONOS_validation"},
        .enabledExtensionCount = 4,
        .ppEnabledExtensionNames =
            (const char*[]){"VK_KHR_surface", "VK_KHR_xcb_surface",
                            VK_EXT_DEBUG_UTILS_EXTENSION_NAME,
                            VK_EXT_DEBUG_REPORT_EXTENSION_NAME}};
    VK_ASSERT(vkCreateInstance(&instanceCreateInfo, nullptr, &instance));

    VkPhysicalDevice physicalDevice = {0};
    VK_ASSERT(findSuitablePhysicalDevice(instance, &physicalDevice));
    VkPhysicalDeviceFeatures deviceFeatures = {0};
    vkGetPhysicalDeviceFeatures(physicalDevice, &deviceFeatures);

    u32 graphicsFamily =
        findQueueFamilies(physicalDevice, VK_QUEUE_GRAPHICS_BIT);
    if(graphicsFamily == (u32)-1) {
        LOG_ERROR("Cannot find queue family");
        return true;
    }
    VkDeviceQueueCreateInfo dqInfo = {
        .sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
        .queueFamilyIndex = graphicsFamily,
        .queueCount = 1,
        .pQueuePriorities = (float[]){1.0f}};
    VkDeviceCreateInfo deviceCreateInfo = {
        .sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO,
        .queueCreateInfoCount = 1,
        .pQueueCreateInfos = &dqInfo,
        .enabledExtensionCount = 1,
        .ppEnabledExtensionNames =
            (const char*[]){VK_KHR_SWAPCHAIN_EXTENSION_NAME},
        .pEnabledFeatures = &deviceFeatures};
    VkDevice device = {0};
    VK_ASSERT(
        vkCreateDevice(physicalDevice, &deviceCreateInfo, nullptr, &device));

    return false;
}

void closeWindow(void) {
    if(window != nullptr) {
        glfwDestroyWindow(window);
        window = nullptr;
    }
    glfwTerminate();
}

void showWindow(void) {
    glfwShowWindow(window);
}

void trapWindowCursor(void) {
    glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
}

void freeWindowCursor(void) {
    glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
}

bool isWindowCursorTrapped(void) {
    return glfwGetInputMode(window, GLFW_CURSOR) == GLFW_CURSOR_DISABLED;
}

const IntVector2* getWindowSize(void) {
    return &size;
}

bool hasWindowSizeChanged(void) {
    return sizeChanged;
}

bool shouldWindowClose(void) {
    return glfwWindowShouldClose(window);
}

void setWindowRunHandler(WindowRunHandler wr, void* data) {
    runHandlerData = data;
    runHandler = wr == nullptr ? dummyWindowRunHandler : wr;
}

void setWindowTickHandler(WindowTickHandler t, void* data) {
    tickHandlerData = data;
    tickHandler = t == nullptr ? dummyWindowTickHandler : t;
}

void setWindowRenderHandler(WindowRenderHandler r, void* data) {
    renderHandlerData = data;
    renderHandler = r == nullptr ? dummyWindowRenderHandler : r;
}

void setWindowNanosPerTick(i64 nanos) {
    nanosPerTick = nanos <= 0 ? 1 : nanos;
}

static i64 updateClock(Clock* c) {
    i64 nanos = getNanos();
    if(nanos < 0) {
        LOG_WARNING("Cannot get nanos, using default");
        nanos = 10000000 + c->last;
    }
    i64 diff = nanos - c->last;
    c->last = nanos;
    c->sum -= c->values[c->index];
    c->values[c->index] = diff;
    c->sum += diff;
    c->index = (c->index + 1) % CLOCK_SIZE;
    return diff;
}

static bool searchForGamepad() {
    if(activeController != -1) {
        return true;
    }
    for(int i = GLFW_JOYSTICK_1; i <= GLFW_JOYSTICK_LAST; i++) {
        if(glfwJoystickIsGamepad(i)) {
            activeController = i;
            return true;
        }
    }
    return false;
}

static void checkGamepad() {
    GLFWgamepadstate state;
    if(!glfwGetGamepadState(activeController, &state)) {
        activeController = -1;
        return;
    }
    for(int i = 0; i <= GLFW_GAMEPAD_BUTTON_LAST; i++) {
        Button b = gamepadToButton[i];
        if(b == 0 || b >= buttonIndex) {
            return;
        }
        if(!lastControllerState.buttons[i] && state.buttons[i]) {
            buttons[b].controllerDown++;
        } else if(lastControllerState.buttons[i] && !state.buttons[i]) {
            buttons[b].controllerUp++;
        }
    }
    lastControllerState = state;
}

static void endFrame(void) {
    sizeChanged = false;
    glfwPollEvents();
    if(searchForGamepad()) {
        checkGamepad();
    }
}

static void tickButton(ButtonData* b) {
    bool down = (b->downEvents > 0) || (b->controllerDown > 0);
    bool up = (b->upEvents == b->downEvents) &&
              (b->controllerUp == b->controllerDown);
    if(b->released) {
        b->downTime = 0;
    }
    b->downTime += down;
    b->released = down && up;

    b->downEvents -= b->upEvents;
    b->upEvents = 0;

    b->controllerDown -= b->controllerUp;
    b->controllerUp = 0;
}

static void tick() {
    updateClock(&tps);
    ButtonData* b = buttons;
    for(size_t i = 0; i < buttonIndex; i++) {
        tickButton(b++);
    }
    tickHandler(tickHandlerData);
    lastMousePosition = mousePosition;
}

void runWindow(void) {
    tps.last = getNanos();
    fps.last = getNanos();
    i64 lag = 0;
    while(runHandler(runHandlerData)) {
        lag += updateClock(&fps);
        while(lag >= nanosPerTick) {
            lag -= nanosPerTick;
            tick();
        }
        renderHandler(renderHandlerData, (float)lag / (float)nanosPerTick);
        endFrame();
    }
}

float getWindowTicksPerSecond(void) {
    return (1000000000.0f * CLOCK_SIZE) / (float)tps.sum;
}

float getWindowFramesPerSecond(void) {
    return (1000000000.0f * CLOCK_SIZE) / (float)fps.sum;
}

void setInputLimit(size_t limit) {
    inputLimit = limit >= INPUT_SIZE ? INPUT_SIZE - 1 : limit;
    if(inputLength > inputLimit) {
        inputLength = inputLimit;
    }
    if(inputCursor > inputLimit) {
        inputCursor = inputLimit;
    }
}

void resetInput() {
    inputLength = 0;
    inputCursor = 0;
}

void enableInput() {
    inputActive = true;
}

void disableInput() {
    inputActive = false;
}

bool isInputEnabled() {
    return inputActive;
}

void fillInput(const char* s) {
    resetInput();
    snprintf(input, inputLength, "%s", s);
}

size_t getInputCursor() {
    return inputCursor;
}

void setInputCursor(size_t index) {
    if(index > inputLength) {
        inputCursor = inputLength;
    } else {
        inputCursor = index;
    }
}

const char* getInputString(void) {
    return input;
}

Button addButton(const char* name) {
    if(buttonIndex >= BUTTONS) {
        return 0;
    }
    Button b = buttonIndex++;
    snprintf(buttons[b].name, sizeof(buttons[b].name), "%s", name);
    return b;
}

void bindKeyToButton(Button b, int key) {
    if(key >= 0 && key < KEYS) {
        keyToButton[key] = b;
    }
}

void bindGamepadToButton(Button b, int gamepadButton) {
    if(gamepadButton >= 0 && gamepadButton < GAMEPAD_BUTTONS) {
        gamepadToButton[gamepadButton] = b;
    }
}

void bindMouseToButton(Button b, int mouseButton) {
    if(mouseButton >= 0 && mouseButton < MOUSE_BUTTONS) {
        mouseToButton[mouseButton] = b;
    }
}

Vector2 getLastMousePosition(void) {
    return lastMousePosition;
}

Vector2 getMousePosition(void) {
    return mousePosition;
}

Vector2 getLeftGamepadAxis(void) {
    return (Vector2){{lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_X],
                      lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_Y]}};
}

Vector2 getRightGamepadAxis(void) {
    return (Vector2){{lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_X],
                      lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_Y]}};
}

float getLeftGamepadTrigger(void) {
    return lastControllerState.axes[GLFW_GAMEPAD_AXIS_LEFT_TRIGGER];
}

float getRightGamepadTrigger(void) {
    return lastControllerState.axes[GLFW_GAMEPAD_AXIS_RIGHT_TRIGGER];
}

static const ButtonData* getButton(Button id) {
    return buttons + (id >= buttonIndex ? 0 : id);
}

bool isButtonDown(Button b) {
    return getButton(b)->downTime > 0;
}

int getButtonDownTime(Button b) {
    return getButton(b)->downTime;
}

bool wasButtonReleased(Button b) {
    return getButton(b)->released;
}

const char* getButtonName(Button b) {
    return getButton(b)->name;
}
